"""
Provides:
  1. Library for testing factor strategies
  2. A lot of different statistical metrics for portfolios
  3. Fancy visualization of results

Source: https://github.com/atomtosov/pqr/tree/dev
"""

from .benchmarks import *
from .factor_models import *
from .factors import *
from .metrics import *
from .portfolios import *
from .preprocessing import *
from .tests import *
from .visualization import *



"""
This module contains instruments for preprocessing the data. There is an idiom
"Garbage In, Garbage Out". So, you should responsibly prepare your data for
work.

You need to make sure that all of your matrices have the same format: their
indices and columns are identical, all values in them are presented as numbers,
missed values are treated carefully, etc. Otherwise, building portfolios or
fitting factor models may be possible, but results will be confusing.
"""


from typing import Iterable, Any, List

import numpy as np
import pandas as pd


__all__ = [
    'correct_matrices',
    'replace_with_nan',
]


def correct_matrices(*matrices: pd.DataFrame) -> List[pd.DataFrame]:
    """
    Corrects indices and columns of passed matrices.

    Parameters
    ----------
    matrices
        Matrices to be corrected.
    """

    # collect all
    all_columns = set()
    all_indices = set()
    for matrix in matrices:
        all_columns |= set(matrix.columns)
        all_indices |= set(matrix.index)

    # find what to drop
    columns_to_drop = set()
    indices_to_drop = set()
    for matrix in matrices:
        columns_to_drop |= (all_columns - set(matrix.columns))
        indices_to_drop |= (all_indices - set(matrix.index))

    # drop and sort
    corrected_matrices = []
    for matrix in matrices:
        # drop and sort columns
        matrix_corrected = matrix.drop(columns_to_drop, axis=1,
                                       errors='ignore')
        matrix_corrected = matrix_corrected[
            sorted(matrix_corrected.columns.values)]
        # drop and sort indices
        matrix_corrected = matrix_corrected.drop(indices_to_drop, axis=0,
                                                 errors='ignore')
        matrix_corrected = matrix_corrected.sort_index()

        corrected_matrices.append(matrix_corrected)

    return corrected_matrices


def replace_with_nan(*matrices: pd.DataFrame,
                     to_replace: Iterable[Any]) -> List[pd.DataFrame]:
    """
    Replaces unusual identifiers for missed values with np.nan.

    Parameters
    ----------
    matrices
        Matrices to be processed.
    to_replace
        Aliases for nans in data.
    """

    replaced_matrices = []
    for matrix in matrices:
        replaced_matrix = matrix.replace(to_replace, np.nan)
        replaced_matrices.append(replaced_matrix)
    return replaced_matrices



"""
This module contains instruments to test, whether performance of a portfolio is
statistically significant or not. The process of statistical testing is
significant part, when developing a portfolio, because it is very easy to fall
into the trap: you can get a portfolio with very high total returns, but these
returns are gotten randomly, so it is not likely that you wanna use this
strategy in future.

For now, it has only 1 test - random test, but more will be added soon.
"""

from typing import Tuple, Callable, Union

import numpy as np
import pandas as pd

import pqr.portfolios

__all__ = [
    'zero_intelligence_test',
]


def zero_intelligence_test(stock_prices: pd.DataFrame,
                           portfolio: pqr.portfolios.Portfolio,
                           target_metric: Callable[
                               [pqr.portfolios.AbstractPortfolio],
                               Union[int, float]],
                           quantiles: int = 10,
                           **kwargs) -> Tuple[pd.Series, np.ndarray]:
    """
    Creates random portfolios, replicating positions of `portfolio`.

    After creating random portfolios, calculates target for each of them. Then,
    selects `n_quantiles` portfolios by quantiles of target.

    Parameters
    ----------
    stock_prices
        Prices, representing stock universe.
    portfolio
        Portfolio, to be tested (replicated by random portfolios).
    target_metric
        Function-like object, computing some number (e.g. value of metric). It
        must get Portfolio and return int or float.
    quantiles
        How many quantile-bounds to generate.
    **kwargs
        Keyword arguments for building up portfolios. See random_portfolios().
    """

    random_portfolios = pqr.portfolios.generate_random_portfolios(
        stock_prices,
        portfolio,
        **kwargs)

    target_values = pd.Series(
        [target_metric(p) for p in random_portfolios])

    indices = []
    for q in np.linspace(0, 1, quantiles):
        indices.append(
            target_values[target_values <= target_values.quantile(q)].argmax())

    target_quantiles = target_values[indices].sort_values()

    return (target_quantiles,
            np.array(random_portfolios)[target_quantiles.index])



"""
This module contains stuff to create portfolios of stocks. Portfolio is a
rebalancing each period batch of stocks. So, portfolio can be represented as
matrix with zeros on cells, where a stock is not picked, and weights (or lots)
on cells, where a stock is picked into a portfolio.

Portfolio can be constructed relatively or with money. Relative portfolio is a
theoretical portfolio, built up with assumption that all available money is
invested in each period (and each period returns are reinvested). Money
portfolio is a more realistic portfolio, which operates with real balance,
which imposes restrictions on the process of investing: practically never
weights of positions are equal to theoretical weights (because of
indivisibility of stocks). Money portfolio is going close to relative one with
initial balance going to infinity.

Also, portfolio can be winners-minus-losers if you construct a factor model.
This type of portfolio is supported, but without real allocation: positions of
winners and losers portfolios a simply used as positions for the wml-portfolio.
Moreover, this portfolio is always theoretical, as it is assumed as
self-financing portfolio (money from shorts are used to long stocks).

Random portfolios to test performance of your portfolio are also supported, but
only for long-only (not wml) portfolios.
"""

from __future__ import annotations

from typing import Protocol, Tuple, Optional, Any, List, Literal

import numpy as np
import pandas as pd

import pqr.factors

__all__ = [
    'AbstractPortfolio',
    'Portfolio',
    'WmlPortfolio',
    'RandomPortfolio',
    'generate_random_portfolios',
]


class AbstractPortfolio(Protocol):
    """
    Abstract class for portfolios.

    Describes, what fields must be included into concrete portfolios and
    realizes methods for fancy printing.
    """

    name: str
    """Name of the portfolio."""
    picks: pd.DataFrame
    """Matrix with stocks picked into the portfolio."""
    weights: pd.DataFrame
    """Matrix with weights in the portfolio for each stock in each period."""
    positions: pd.DataFrame
    """Matrix with allocated stocks into the portfolio in each period."""
    returns: pd.Series
    """Periodical returns of the portfolio (non-cumulative)."""

    def __repr__(self) -> str:
        return f'{type(self).__name__}({repr(self.name)})'

    def __str__(self) -> str:
        return self.name


class Portfolio(AbstractPortfolio):
    """
    Class for factor portfolios.

    Parameters
    ----------
    name
        Name of the portfolio.
    """

    def __init__(self, name: str = 'portfolio'):
        self.name = name

        self.picks = pd.DataFrame()
        self.weights = pd.DataFrame()
        self.positions = pd.DataFrame()
        self.returns = pd.Series()

    def pick_stocks_by_factor(
            self,
            factor: pqr.factors.Factor,
            thresholds: Tuple[int | float, int | float],
            method: Literal['quantile', 'top', 'time-series'] = 'quantile'
    ) -> Portfolio:
        """
        Picks subset of stocks into the portfolio, choosing them by `factor`.

        Supports 3 methods to pick stocks:

        * quantile
        * top
        * time-series

        Parameters
        ----------
        factor
            Factor to pick stocks into the portfolio.
        thresholds
            Bounds for the set of allowed values of `factor` to pick stocks.
        method
            Method, used to define subset of stocks to be picked on the basis
            of given `thresholds`.
        """

        factor = factor.data

        if method == 'quantile':
            lower_threshold, upper_threshold = np.nanquantile(
                factor, thresholds, axis=1, keepdims=True
            )
        elif method == 'top':
            lower_threshold = np.nanmin(
                factor.apply(pd.Series.nlargest, n=thresholds[1], axis=1),
                axis=1, keepdims=True
            )
            upper_threshold = np.nanmin(
                factor.apply(pd.Series.nlargest, n=thresholds[0], axis=1),
                axis=1, keepdims=True
            )
        else:  # method = 'time-series'
            lower_threshold, upper_threshold = thresholds

        self.picks = (lower_threshold <= factor) & (factor <= upper_threshold)

        return self

    def weigh_by_factor(
            self,
            factor: pqr.factors.Factor
    ) -> Portfolio:
        """
        Weighs the `picks` by `factor`.

        Finds linear weights: simply divides each value in a row by the sum of
        the row. If factor is lower_better (`is_bigger_better` = False), then
        weights are additionally "mirrored".

        Parameters
        ----------
        factor
            Factor to weigh picks.

        Notes
        -----
        Works only for factors with all positive values.
        """

        raw_weights = self.picks * factor.data.loc[self.picks.index[0]:]
        normalizer = np.nansum(raw_weights, axis=1, keepdims=True)

        self.weights = (raw_weights / normalizer).fillna(0)

        return self

    def weigh_equally(self) -> Portfolio:
        """
        Weighs the `picks` equally: all stocks will have the same weights.
        """

        raw_weights = self.picks * np.ones_like(self.picks, dtype=int)
        normalizer = np.nansum(raw_weights, axis=1, keepdims=True)
        self.weights = (raw_weights / normalizer).fillna(0)

        return self

    def scale_weights_by_factor(
            self,
            factor: pqr.factors.Factor,
            target: int | float = 1
    ) -> Portfolio:
        """
        Scale the `weights` by `target` of `factor`.

        Simply divides each value in a row by `target`. If `factor` is
        lower_better (`bigger_better` = False), then leverages is additionally
        "mirrored".

        Parameters
        ----------
        factor
            Factor to scale (leverage) positions weights.
        target
            Target of `factor`.

        Notes
        -----
        Works only for factors with all positive values.
        """

        leveraged_weights = self.weights * factor.data / target

        self.weights = leveraged_weights.fillna(0)

        return self

    def allocate(
            self,
            stock_prices: pd.DataFrame,
            balance: Optional[int | float] = None,
            fee_rate: int | float = 0
    ) -> Portfolio:
        """
        Allocates positions, based on `weights`.

        If `balance` is None:
            positions will be equal to weights with correction on commission.

        If `balance` is number > 0:
            positions will be allocated, using simple greedy algorithm - buy as
            much stocks as possible to be maximum closely to expected weights.

        Parameters
        ----------
        stock_prices
            Prices, representing stock universe.
        balance
            Initial balance of the portfolio.
        fee_rate
            Indicative commission rate for every deal.

        Notes
        -----
        For now allocation with money can lead to negative cash if
        `fee_rate` != 0, but it will be fixed soon.
        """

        if balance is None:
            self._allocate_relatively(stock_prices, fee_rate)
        else:
            self._allocate_with_money(stock_prices, balance, fee_rate)

        self.returns.name = self.name

        return self

    def _allocate_relatively(
            self,
            stock_prices: pd.DataFrame,
            fee_rate: int | float
    ) -> None:
        """
        Allocates positions relatively (theoretically).
        """

        stock_prices = stock_prices.loc[self.weights.index[0]:]
        self.positions = self.weights
        universe_returns = stock_prices.pct_change().shift(-1)
        portfolio_returns = (
            (self.weights * universe_returns).shift().sum(axis=1))

        self.returns = portfolio_returns * (1 - fee_rate)

    def _allocate_with_money(
            self,
            stock_prices: pd.DataFrame,
            balance: int | float,
            fee_rate: int | float
    ) -> None:
        """
        Allocates positions with money.
        """

        stock_prices = stock_prices.loc[self.weights.index[0]:]

        portfolio = self.weights.copy()
        returns = pd.Series(index=portfolio.index)

        portfolio.values[0] = (
                portfolio.values[0] * balance // stock_prices.values[0])
        returns.values[0] = 0
        cash = balance - np.nansum(portfolio.values[0] *
                                   stock_prices.values[0] * (1 + fee_rate))
        prev_balance = balance
        for i in range(1, len(portfolio)):
            w = portfolio.values[i]
            prices = stock_prices.values[i]
            prev_alloc = portfolio.values[i - 1]

            cur_balance = cash + np.nansum(prev_alloc * prices)
            alloc = w * cur_balance // prices

            alloc_diff = alloc - prev_alloc
            cash_diff = -(alloc_diff * prices)
            commission = np.abs(cash_diff) * fee_rate

            cash += np.nansum(cash_diff - commission)

            portfolio.values[i] = alloc
            returns.values[i] = cur_balance / prev_balance - 1

            prev_balance = cur_balance

        self.positions = portfolio
        self.returns = returns


class WmlPortfolio(AbstractPortfolio):
    """
    Class for WML (winners-minus-losers) portfolios.

    This type of portfolio is always treated as theoretical (self-financing),
    and that is why it assumes that given portfolios are relevant: they are
    built with the same balance, or both of them are relative. Actually, it
    simply subtracts `losers_portfolio` from `winners_portfolio`.

    Parameters
    ----------
    winners_portfolio
        Portfolio of winners by some factor. Positions of this portfolio will
        be long-positions for a wml-portfolio.
    losers_portfolio
        Portfolio of losers by the same factor. Positions of this portfolio
        will be short-positions for a wml-portfolio.
    name
        Name of the wml-portfolio.
    """

    def __init__(
            self,
            winners_portfolio: Portfolio,
            losers_portfolio: Portfolio,
            name: str = 'wml'
    ):
        self.name = name

        self.picks = (winners_portfolio.picks.astype(int) -
                      losers_portfolio.picks.astype(int))
        self.weights = winners_portfolio.weights - losers_portfolio.weights
        self.positions = (winners_portfolio.positions -
                          losers_portfolio.positions)
        self.returns = winners_portfolio.returns - losers_portfolio.returns
        self.returns.name = self.name


class RandomPortfolio(AbstractPortfolio):
    """
    Class for random portfolios, replicating picks of a portfolio.

    It implements the delegation pattern, so it has the same interface as
    Portfolio with the only one exception: the method
    ``pick_stocks_by_factor()`` is replaced with the method
    ``pick_stocks_randomly()``.

    Parameters
    ----------
    name
        Name of the random portfolio.
    random_seed
        Random seed to make random deterministic.
    """

    def __init__(
            self,
            name: str = 'random',
            random_seed: Optional[int] = None
    ):
        self._portfolio = Portfolio(name)

        if random_seed is not None:
            np.random.seed(random_seed)

    def __getattr__(self, name: str) -> Any:
        if name == 'pick_stocks_by_factor':
            raise AttributeError('\'RandomPortfolio\' object has no attribute '
                                 '\'pick_stocks_by_factor\'')
        return getattr(self._portfolio, name)

    def pick_stocks_randomly(
            self,
            picks: pd.DataFrame,
            mask: Optional[pd.DataFrame] = None
    ) -> RandomPortfolio:
        """
        Pick stocks randomly, but in the same quantity as in the `picks`.

        In each period collects number of picked stocks and randomly pick the
        same amount of stocks into a random portfolio.

        Parameters
        ----------
        picks
            Picks to replicate by the random portfolio.
        mask
            Matrix to prohibit pick stock during periods, when they are not
            really available for trading. Also can be used to filter stock
            universe.
        """

        picks = picks.copy().astype(float)
        if mask is not None:
            picks[~mask] = np.nan

        def random_pick(row: np.ndarray,
                        indices=np.indices((picks.shape[1],))[0]):
            picked = (row > 0).sum()
            choice = np.random.choice(indices[~np.isnan(row)], picked)
            random_picks = np.zeros_like(row, dtype=bool)
            random_picks[choice] = True
            return random_picks

        self._portfolio.picks = pd.DataFrame(
            np.apply_along_axis(random_pick, axis=1, arr=picks.values),
            index=picks.index, columns=picks.columns
        )

        return self

    def weigh_by_factor(
            self,
            factor: pqr.factors.Factor
    ) -> RandomPortfolio:
        """
        See Portfolio.weigh_by_factor().
        """

        self._portfolio.weigh_by_factor(factor)

        return self

    def weigh_equally(self) -> RandomPortfolio:
        """
        See Portfolio.weigh_equally().
        """

        self._portfolio.weigh_equally()

        return self

    def scale_weights_by_factor(
            self,
            factor: pqr.factors.Factor,
            target: int | float = 1
    ) -> RandomPortfolio:
        """
        See Portfolio.scale_weights_by_factor().
        """

        self._portfolio.scale_weights_by_factor(factor, target)

        return self

    def allocate(
            self,
            stock_prices: pd.DataFrame,
            balance: Optional[int | float] = None,
            fee_rate: int | float = 0
    ) -> RandomPortfolio:
        """
        See Portfolio.allocate().
        """

        self._portfolio.allocate(stock_prices, balance, fee_rate)

        return self


def generate_random_portfolios(
        stock_prices: pd.DataFrame,
        portfolio: Portfolio,
        mask: Optional[pd.DataFrame] = None,
        weighting_factor: Optional[pqr.factors.Factor] = None,
        scaling_factor: Optional[pqr.factors.Factor] = None,
        target: int | float = 1,
        balance: Optional[int | float] = None,
        fee_rate: int | float = 0,
        n: int = 100,
        random_seed: Optional[int] = None
) -> List[RandomPortfolio]:
    """
    Creates `n` random portfolios, replicating portfolio `picks`.

    Parameters
    ----------
    stock_prices
        Prices, representing stock universe.
    portfolio
        Portfolio to be replicated by random ones.
    mask
        Matrix to filter stock universe. By default, stocks which are not
        available for trading are excluded, so there is no necessity to pass
        this mask, if specific filter is not used.
    weighting_factor
        Factor, weighting positions. If not given, just weigh equally.
    scaling_factor
        Factor to scale (leverage) positions weights.
    target
        Target of `scaling_factor`.
    balance
        Initial balance of the portfolio.
    fee_rate
        Indicative commission rate for every deal.
    n
        How many random portfolios to be created.
    random_seed
        Random seed to make random deterministic.
    """

    np.random.seed(random_seed)

    picks = portfolio.picks.astype(float)
    picks[stock_prices.isna()] = np.nan
    if mask is not None:
        picks[~mask] = np.nan

    portfolios = []
    for i in range(n):
        random_portfolio = RandomPortfolio()
        random_portfolio.pick_stocks_randomly(picks)
        if weighting_factor is not None:
            random_portfolio.weigh_by_factor(weighting_factor)
        else:
            random_portfolio.weigh_equally()
        if scaling_factor is not None:
            random_portfolio.scale_weights_by_factor(scaling_factor, target)
        random_portfolio.allocate(stock_prices, balance, fee_rate)

        portfolios.append(random_portfolio)

    return portfolios


def _ranked_reverse(data: pd.DataFrame) -> pd.DataFrame:
    """
    Algorithm, mirroring weights for lower_better factors.
    """

    data = data.copy()
    straight_sort = np.argsort(data.values, axis=1)
    reversed_sort = np.fliplr(straight_sort)
    for i in range(len(data)):
        data.values[i, straight_sort[i]] = data.values[i, reversed_sort[i]]

    return data



"""
This module contains all necessary graphics to visualize. Visualization is very
important while assessing performance of a portfolio or a factor model. Some
features or simply errors can be found only by eye.

For now it includes only very simple visualization of results, but it will be
developed in future.
"""

from typing import Optional

import matplotlib.pyplot as plt

__all__ = [
    'plot_cumulative_returns',
]

import pqr.metrics
import pqr.portfolios
import pqr.benchmarks


def plot_cumulative_returns(
        *portfolios: pqr.portfolios.AbstractPortfolio,
        benchmark: Optional[pqr.benchmarks.Benchmark] = None
) -> None:
    """
    Plots cumulative returns of portfolios (optionally with a benchmark).

    Parameters
    ----------
    portfolios
        Portfolios, which cumulative returns are plotted.
    benchmark
        Benchmark or portfolio to plot cumulative returns "reference point".
    """

    for portfolio in portfolios:
        pqr.metrics.cumulative_returns(portfolio).plot()

    if benchmark is not None:
        start_trading = portfolios[0].returns.index[0]
        benchmark_cum_returns = pqr.metrics.cumulative_returns(benchmark)
        benchmark_cum_returns = (benchmark_cum_returns[start_trading:] -
                                 benchmark_cum_returns[start_trading])
        benchmark_cum_returns.plot(color='gray', alpha=0.8)

    plt.title('Portfolio Cumulative Returns', fontsize=25)
    plt.grid()
    plt.legend()



"""
This module contains metrics and statistics to assess performance of a
portfolio. Usual metrics are always numbers (int or float), but also rolling
metrics are supported. Rolling metrics are calculated to estimate annual
performance of a portfolio in each trading period: every period gather not all
points of returns but only for 1 year (e.g. if returns are monthly, rolling
window size is 12).

For now practically all popular in portfolio management metrics and statistics
are supported, but you are welcome to create your own metrics and contribute to
source code.
"""

from __future__ import annotations

from typing import Callable, Optional

import numpy as np
import pandas as pd
import statsmodels.regression.linear_model as sm_linear
import statsmodels.tools.tools as sm_tools

import pqr.benchmarks
import pqr.portfolios

__all__ = [
    'summary',

    'cumulative_returns', 'total_return',
    'annual_return', 'annual_volatility',

    'mean_return', 'rolling_mean_return',
    'volatility', 'rolling_volatility',
    'max_drawdown', 'rolling_max_drawdown',
    'win_rate', 'rolling_win_rate',

    'value_at_risk', 'rolling_value_at_risk',
    'expected_tail_loss', 'rolling_expected_tail_loss',
    'rachev_ratio', 'rolling_rachev_ratio',

    'calmar_ratio', 'rolling_calmar_ratio',
    'sharpe_ratio', 'rolling_sharpe_ratio',
    'omega_ratio', 'rolling_omega_ratio',
    'sortino_ratio', 'rolling_sortino_ratio',
    'downside_risk', 'rolling_downside_risk',

    'mean_excess_return', 'rolling_mean_excess_return',
    'benchmark_correlation', 'rolling_benchmark_correlation',
    'alpha', 'rolling_alpha',
    'beta', 'rolling_beta',
]


def summary(
        portfolio: pqr.portfolios.AbstractPortfolio,
        benchmark: pqr.benchmarks.Benchmark
) -> pd.Series:
    """
    Calculates summary statistics for a `portfolio`.

    Computed metrics:

    - Alpha, %
    - Beta, %
    - Sharpe Ratio
    - Mean Return, %
    - Volatility, %
    - Excess Return, %
    - Benchmark Correlation, %
    - Win Rate, %
    - Maximum Drawdown, %

    Parameters
    ----------
    portfolio
        Portfolio, for which metrics are calculated.
    benchmark
        Benchmark, which used as the alternative for the `portfolio` to
        calculate some metrics (usually some stock index).
    """

    return pd.Series(
        {
            'Alpha, %': alpha(portfolio, benchmark) * 100,
            'Beta': beta(portfolio, benchmark),
            'Sharpe Ratio': sharpe_ratio(portfolio),
            'Mean Return, %': mean_return(portfolio) * 100,
            'Volatility, %': volatility(portfolio) * 100,
            'Mean Excess Return, %': mean_excess_return(portfolio,
                                                        benchmark) * 100,
            'Benchmark Correlation': benchmark_correlation(portfolio,
                                                           benchmark),
            'Win Rate, %': win_rate(portfolio) * 100,
            'Maximum Drawdown, %': max_drawdown(portfolio) * 100,
        },
        name=portfolio.name
    ).round(2)


def cumulative_returns(
        portfolio: pqr.portfolios.AbstractPortfolio | pqr.benchmarks.Benchmark
) -> pd.Series:
    """
    Calculates Cumulative Returns of a `portfolio`.

    Parameters
    ----------
    portfolio
        An allocated portfolio or a benchmark.

    Notes
    -----
    Starting point equals to zero.
    """

    return _cumulative_returns(portfolio.returns)


def total_return(portfolio: pqr.portfolios.AbstractPortfolio) -> int | float:
    """
    Calculates Total Return of a `portfolio`.

    Parameters
    ----------
    portfolio
        An allocated portfolio.

    Notes
    -----
    Shows additional return, not equity curve's state.
    """

    return _total_return(portfolio.returns)


def annual_return(portfolio: pqr.portfolios.AbstractPortfolio) -> int | float:
    """
    Calculates Annual Return of a `portfolio` as CAGR.

    CAGR (Compounded Annual Growth Rate) calculated as:

    .. math::
        CAGR = (1 + Total Return)^{\\frac{1}{Years}} - 1

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    """

    return _annual_return(portfolio.returns)


def annual_volatility(
        portfolio: pqr.portfolios.AbstractPortfolio) -> int | float:
    """
    Calculates Annual Volatility of returns of a `portfolio`.

    Annual Volatility of the portfolio is the annualized standard deviation of
    portfolio returns:

    .. math::
        \\sigma_r * \\sqrt{Number\;of\;Periods\;in\;a\;Year}

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    """

    return _annual_volatility(portfolio.returns)


def mean_return(portfolio: pqr.portfolios.AbstractPortfolio) -> int | float:
    """
    Calculates Mean Return of a `portfolio`.

    Mean Return is simple expected value:

    .. math::
        E(r) = \\bar{r}

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    """

    return _mean_return(portfolio.returns)


def rolling_mean_return(
        portfolio: pqr.portfolios.AbstractPortfolio,
        window: Optional[int] = None
) -> pd.Series:
    """
    Calculates rolling Mean Return of a `portfolio`.

    See :func:`~pqr.metrics.mean_return`.

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    window
        Number of observations on a rolling window. If not passed, `window`
        equals to approximate number of periods in a year.
    """

    return _roll(
        portfolio.returns,
        metric=_mean_return,
        window=window
    )


def volatility(portfolio: pqr.portfolios.AbstractPortfolio) -> int | float:
    """
    Calculates Volatility of returns of a `portfolio`.

    Volatility of the portfolio is the standard deviation of portfolio returns:

    .. math::
        \\sigma_r = \\sqrt{\\frac{\\sum_{i=1}^{n}(r_i-\\bar{r})^2}{n-1}}

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    """

    return _volatility(portfolio.returns)


def rolling_volatility(
        portfolio: pqr.portfolios.AbstractPortfolio,
        window: Optional[int] = None
) -> pd.Series:
    """
    Calculates rolling Volatility of a `portfolio`.

    See :func:`~pqr.metrics.volatility`.

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    window
        Number of observations on a rolling window. If not passed, `window`
        equals to approximate number of periods in a year.
    """

    return _roll(
        portfolio.returns,
        metric=_volatility,
        window=window
    )


def max_drawdown(portfolio: pqr.portfolios.AbstractPortfolio) -> int | float:
    """
    Calculates Maximum Drawdown of a `portfolio`.

    Maximum Drawdown of portfolio is the highest relative difference between
    high water mark (cumulative maximum of the cumulative returns) and
    cumulative returns:

    .. math::
        MDD = -\max{\\frac{HWM - Cumulative\;Returns}{HWM}}

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    """

    return _max_drawdown(portfolio.returns)


def rolling_max_drawdown(
        portfolio: pqr.portfolios.AbstractPortfolio,
        window: Optional[int] = None
) -> pd.Series:
    """
    Calculates rolling Maximum Drawdown of a `portfolio`.

    See :func:`~pqr.metrics.max_drawdown`.

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    window
        Number of observations on a rolling window. If not passed, `window`
        equals to approximate number of periods in a year.
    """

    return _roll(
        portfolio.returns,
        metric=_max_drawdown,
        window=window
    )


def win_rate(portfolio: pqr.portfolios.AbstractPortfolio) -> int | float:
    """
    Calculates Win Rate of a `portfolio`.

    Win Rate of a portfolio is simple ratio of number of periods with positive
    returns and total number of trading periods:

    .. math::
        WR = \\frac{\\sum_{i=1}^{n}[r_{i} > 0]}{n}

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    """

    return _win_rate(portfolio.returns)


def rolling_win_rate(
        portfolio: pqr.portfolios.AbstractPortfolio,
        window: Optional[int] = None
) -> pd.Series:
    """
    Calculates rolling Win Rate of a `portfolio`.

    See :func:`~pqr.metrics.win_rate`.

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    window
        Number of observations on a rolling window. If not passed, `window`
        equals to approximate number of periods in a year.
    """

    return _roll(
        portfolio.returns,
        metric=_win_rate,
        window=window
    )


def value_at_risk(
        portfolio: pqr.portfolios.AbstractPortfolio,
        confidence_level: int | float = 0.95
) -> int | float:
    """
    Calculates Value at Risk of a `portfolio`.

    VaR shows the amount of potential loss that could happen in a portfolio
    with given `confidence_level`:

    .. math::
        VaR = -\\inf\{F_r(r) > Confidence\;Level\}

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    confidence_level
        The probability, with which the estimation of VaR is true.
    """

    return _value_at_risk(portfolio.returns, confidence_level)


def rolling_value_at_risk(
        portfolio: pqr.portfolios.AbstractPortfolio,
        confidence_level: int | float = 0.95,
        window: Optional[int] = None
) -> pd.Series:
    """
    Calculates rolling Value at Risk of a `portfolio`.

    See :func:`~pqr.metrics.value_at_risk`.

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    confidence_level
        The probability, with which the estimation of VaR is true.
    window
        Number of observations on a rolling window. If not passed, `window`
        equals to approximate number of periods in a year.
    """

    return _roll(
        portfolio.returns,
        metric=_value_at_risk,
        window=window,
        confidence_level=confidence_level
    )


def expected_tail_loss(
        portfolio: pqr.portfolios.AbstractPortfolio,
        confidence_level: int | float = 0.95
) -> int | float:
    """
    Calculates Expected Tail Loss of a `portfolio`.

    Expected Tail Loss shows the average of the values that fall beyond the
    VaR, calculated with `confidence_level`:

    .. math::
        ETL = \\frac{\\sum_{i=1}^{n}r_i\cdot[r_i \le VaR]}
        {\\sum_{i=1}^{n}[r_i \le VaR]}

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    confidence_level
        The probability, with which the estimation of VaR is true.
    """

    return _expected_tail_loss(portfolio.returns, confidence_level)


def rolling_expected_tail_loss(
        portfolio: pqr.portfolios.AbstractPortfolio,
        confidence_level: int | float = 0.95,
        window: Optional[int] = None
) -> pd.Series:
    """
    Calculates rolling Expected Tail Loss of a `portfolio`.

    See :func:`~pqr.metrics.expected_tail_loss`.

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    confidence_level
        The probability, with which the estimation of VaR is true.
    window
        Number of observations on a rolling window. If not passed, `window`
        equals to approximate number of periods in a year.
    """

    return _roll(
        portfolio.returns,
        metric=_expected_tail_loss,
        window=window,
        confidence_level=confidence_level
    )


def rachev_ratio(
        portfolio: pqr.portfolios.AbstractPortfolio,
        reward_cutoff: int | float = 0.95,
        risk_cutoff: int | float = 0.05,
        risk_free_rate: int | float | pd.Series = 0
) -> int | float:
    """
    Calculates Rachev Ratio (R-Ratio) of a `portfolio`.

    Rachev Ratio calculated as ETR (Expected Tail Return) divided by ETL
    (Expected Tail Return) of adjusted by `risk_free_rate` returns:

    .. math::
        R-Ratio = \\frac{ETR(r)_\\alpha}{ETL(r)_\\beta}

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    reward_cutoff
        Cutoff to calculate expected tail return.
    risk_cutoff
        Cutoff to calculate expected tail loss. Confidence level to compute it
        equals to 1 - `risk_cutoff`.
    risk_free_rate
        Indicative rate of guaranteed returns (e.g. US government bond rate).
    """

    return _rachev_ratio(portfolio.returns, reward_cutoff, risk_cutoff,
                         risk_free_rate)


def rolling_rachev_ratio(
        portfolio: pqr.portfolios.AbstractPortfolio,
        reward_cutoff: int | float = 0.95,
        risk_cutoff: int | float = 0.05,
        risk_free_rate: int | float | pd.Series = 0,
        window: Optional[int] = None
) -> pd.Series:
    """
    Calculates rolling Rachev Ratio (R-Ratio) of a `portfolio`.

    See :func:`~pqr.metrics.rachev_ratio`.

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    reward_cutoff
        Cutoff to calculate expected tail return.
    risk_cutoff
        Cutoff to calculate expected tail loss. Confidence level to compute it
        equals to 1 - `risk_cutoff`.
    risk_free_rate
        Indicative rate of guaranteed returns (e.g. US government bond rate).
    window
        Number of observations on a rolling window. If not passed, `window`
        equals to approximate number of periods in a year.
    """

    return _roll(
        portfolio.returns,
        metric=_mean_return,
        window=window,
        reward_cutoff=reward_cutoff,
        risk_cutoff=risk_cutoff,
        risk_free_rate=risk_free_rate
    )


def calmar_ratio(portfolio: pqr.portfolios.AbstractPortfolio) -> int | float:
    """
    Calculates Calmar Ratio of a `portfolio`.

    Calmar Ratio is annual return (CAGR) divided by maximum drawdown of the
    period:

    .. math::
        CR = \\frac{CAGR(r)}{MDD(r)}

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    """

    return _calmar_ratio(portfolio.returns)


def rolling_calmar_ratio(
        portfolio: pqr.portfolios.AbstractPortfolio,
        window: Optional[int] = None
) -> pd.Series:
    """
    Calculates rolling Calmar Ratio of a `portfolio`.

    See :func:`~pqr.metrics.calmar_ratio`.

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    window
        Number of observations on a rolling window. If not passed, `window`
        equals to approximate number of periods in a year.
    """

    return _roll(
        portfolio.returns,
        metric=_calmar_ratio,
        window=window
    )


def sharpe_ratio(
        portfolio: pqr.portfolios.AbstractPortfolio,
        risk_free_rate: int | float | pd.Series = 0
) -> int | float:
    """
    Calculates Sharpe Ratio of a `portfolio`.

    Sharpe Ratio calculated as annualized ratio between mean and volatility of
    adjusted by `risk_free_rate` returns:

    .. math::
        SR = \\frac{\\bar{r}}{\\sigma_{r}}

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    risk_free_rate
        Indicative rate of guaranteed returns (e.g. US government bond rate).
    """

    return _sharpe_ratio(portfolio.returns, risk_free_rate)


def rolling_sharpe_ratio(
        portfolio: pqr.portfolios.AbstractPortfolio,
        risk_free_rate: int | float | pd.Series = 0,
        window: Optional[int] = None
) -> pd.Series:
    """
    Calculates rolling Sharpe Ratio of a `portfolio`.

    See :func:`~pqr.metrics.sharpe_ratio`.

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    risk_free_rate
        Indicative rate of guaranteed returns (e.g. US government bond rate).
    window
        Number of observations on a rolling window. If not passed, `window`
        equals to approximate number of periods in a year.
    """

    return _roll(
        portfolio.returns,
        metric=_sharpe_ratio,
        window=window,
        risk_free_rate=risk_free_rate
    )


def omega_ratio(
        portfolio: pqr.portfolios.AbstractPortfolio,
        required_return: int | float | pd.Series = 0
) -> int | float:
    """
    Calculates Omega Ratio of a `portfolio`.

    Omega Ratio calculated as the area of the probability distribution function
    of returns above `required_return` divided by the area under
    `required_return`:

    .. math::
        \\Omega(\\theta) = \\frac{\\int_{\\theta}^{\\infty}[1-F(r)]dr}
        {\\int_{-\\infty}^{\\theta}F(r)dr}

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    required_return
        Rate of returns, required for an investor.
    """

    return _omega_ratio(portfolio.returns, required_return)


def rolling_omega_ratio(
        portfolio: pqr.portfolios.AbstractPortfolio,
        required_return: int | float | pd.Series = 0,
        window: Optional[int] = None
) -> pd.Series:
    """
    Calculates rolling Omega Ratio of a `portfolio`.

    See :func:`~pqr.metrics.omega_ratio`.

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    required_return
        Rate of returns, required for an investor.
    window
        Number of observations on a rolling window. If not passed, `window`
        equals to approximate number of periods in a year.
    """

    return _roll(
        portfolio.returns,
        metric=_omega_ratio,
        window=window,
        required_return=required_return
    )


def sortino_ratio(
        portfolio: pqr.portfolios.AbstractPortfolio,
        minimum_acceptable_return: int | float | pd.Series = 0
) -> int | float:
    """
    Calculates Sortino Ratio of a `portfolio`.

    Sortino Ratio is the mean of adjusted by `minimum_acceptable_return`
    `portfolio` returns divided by Downside Risk:

    .. math::
        SR = \\frac{\\bar{r}}{Downside\;Risk}

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    minimum_acceptable_return
        Rate of minimum acceptable returns, required for an investor.
    """

    return _sortino_ratio(portfolio.returns, minimum_acceptable_return)


def rolling_sortino_ratio(
        portfolio: pqr.portfolios.AbstractPortfolio,
        minimum_acceptable_return: int | float | pd.Series = 0,
        window: Optional[int] = None
) -> pd.Series:
    """
    Calculates rolling Sortino Ratio of a `portfolio`.

    See :func:`~pqr.metrics.sortino_ratio`.

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    minimum_acceptable_return
        Rate of returns, required for an investor.
    window
        Number of observations on a rolling window. If not passed, `window`
        equals to approximate number of periods in a year.
    """

    return _roll(
        portfolio.returns,
        metric=_sortino_ratio,
        window=window,
        minimum_acceptable_return=minimum_acceptable_return
    )


def downside_risk(
        portfolio: pqr.portfolios.AbstractPortfolio,
        minimum_acceptable_return: int | float | pd.Series
) -> int | float:
    """
    Calculates Downside Risk of a `portfolio`.

    Downside Risk is the annualized deviation of `portfolio` returns under
    `minimum_acceptable_return`:

    .. math::
        DR = \\sqrt{\\frac{\\sum_{i=1}^{n}max(r_i-mar, 0)}{n}}

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    minimum_acceptable_return
        Rate of minimum acceptable returns, required for an investor.
    """

    return _downside_risk(portfolio.returns, minimum_acceptable_return)


def rolling_downside_risk(
        portfolio: pqr.portfolios.AbstractPortfolio,
        minimum_acceptable_return: int | float | pd.Series = 0,
        window: Optional[int] = None
) -> pd.Series:
    """
    Calculates rolling Downside Risk of a `portfolio`.

    See :func:`~pqr.metrics.downside_risk`.

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    minimum_acceptable_return
        Rate of returns, required for an investor.
    window
        Number of observations on a rolling window. If not passed, `window`
        equals to approximate number of periods in a year.
    """

    return _roll(
        portfolio.returns,
        metric=_downside_risk,
        window=window,
        minimum_acceptable_return=minimum_acceptable_return
    )


def mean_excess_return(
        portfolio: pqr.portfolios.AbstractPortfolio,
        benchmark: pqr.benchmarks.Benchmark
) -> int | float:
    """
    Calculates Mean Excess Return of a `portfolio`.

    Mean Excess Return is the mean difference between `portfolio` returns and
    `benchmark` returns:

    .. math::
        MER = E(r_{portfolio} - r_{benchmark})

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    benchmark
        Benchmark for the `portfolio` to beat.
    """

    return _mean_excess_return(portfolio.returns, benchmark.returns)


def rolling_mean_excess_return(
        portfolio: pqr.portfolios.AbstractPortfolio,
        benchmark: pqr.benchmarks.Benchmark,
        window: Optional[int] = None
) -> pd.Series:
    """
    Calculates rolling Mean Excess Return of a `portfolio`.

    See :func:`~pqr.metrics.mean_excess_return`.

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    benchmark
        Benchmark for the `portfolio` to beat.
    window
        Number of observations on a rolling window. If not passed, `window`
        equals to approximate number of periods in a year.
    """

    return _roll(
        portfolio.returns,
        benchmark.returns,
        metric=_mean_excess_return,
        window=window
    )


def benchmark_correlation(
        portfolio: pqr.portfolios.AbstractPortfolio,
        benchmark: pqr.benchmarks.Benchmark
) -> int | float:
    """
    Calculates Benchmark Correlation of a `portfolio`.

    Benchmark Correlation is the simple spearman correlation between
    `portfolio` returns and `benchmark` returns:

    .. math::
        BC = corr(r_{portfolio}, r_{benchmark})

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    benchmark
        Benchmark for the `portfolio` to beat.
    """

    return _benchmark_correlation(portfolio.returns, benchmark.returns)


def rolling_benchmark_correlation(
        portfolio: pqr.portfolios.AbstractPortfolio,
        benchmark: pqr.benchmarks.Benchmark,
        window: Optional[int] = None
) -> pd.Series:
    """
    Calculates rolling Benchmark Correlation of a `portfolio`.

    See :func:`~pqr.metrics.benchmark_correlation`.

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    benchmark
        Benchmark for the `portfolio` to beat.
    window
        Number of observations on a rolling window. If not passed, `window`
        equals to approximate number of periods in a year.
    """

    return _roll(
        portfolio.returns,
        benchmark.returns,
        metric=_benchmark_correlation,
        window=window
    )


def alpha(
        portfolio: pqr.portfolios.AbstractPortfolio,
        benchmark: pqr.benchmarks.Benchmark,
        risk_free_rate: int | float | pd.Series = 0
) -> int | float:
    """
    Calculates Alpha of a `portfolio`.

    Alpha is the coefficient :math:`\\alpha` in the estimated regression of
    `portfolio` returns per `benchmark` returns (both are adjusted by
    `risk_free_rate`):

    .. math::
        r_{portfolio}-r_f = \\alpha +\\beta*(r_{benchmark}-r_f)+\\epsilon

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    benchmark
        Benchmark for the `portfolio` to beat.
    risk_free_rate
        Indicative rate of guaranteed returns (e.g. US government bond rate).
    """

    return _alpha(portfolio.returns, benchmark.returns, risk_free_rate)


def rolling_alpha(
        portfolio: pqr.portfolios.AbstractPortfolio,
        benchmark: pqr.benchmarks.Benchmark,
        risk_free_rate: int | float | pd.Series = 0,
        window: Optional[int] = None
) -> pd.Series:
    """
    Calculates rolling Alpha of a `portfolio`.

    See :func:`~pqr.metrics.alpha`.

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    benchmark
        Benchmark for the `portfolio` to beat.
    risk_free_rate
        Indicative rate of guaranteed returns (e.g. US government bond rate).
    window
        Number of observations on a rolling window. If not passed, `window`
        equals to approximate number of periods in a year.
    """

    return _roll(
        portfolio.returns,
        benchmark.returns,
        metric=_alpha,
        window=window,
        risk_free_rate=risk_free_rate
    )


def beta(
        portfolio: pqr.portfolios.AbstractPortfolio,
        benchmark: pqr.benchmarks.Benchmark,
        risk_free_rate: int | float | pd.Series = 0
) -> int | float:
    """
    Calculates Beta of a `portfolio`.

    Beta is the coefficient :math:`\\beta` in the estimated regression of
    `portfolio` returns per `benchmark` returns (both are adjusted by
    `risk_free_rate`):

    .. math::
        r_{portfolio}-r_f = \\alpha +\\beta*(r_{benchmark}-r_f)+\\epsilon

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    benchmark
        Benchmark for the `portfolio` to beat.
    risk_free_rate
        Indicative rate of guaranteed returns (e.g. US government bond rate).
    """

    return _beta(portfolio.returns, benchmark.returns, risk_free_rate)


def rolling_beta(
        portfolio: pqr.portfolios.AbstractPortfolio,
        benchmark: pqr.benchmarks.Benchmark,
        risk_free_rate: int | float | pd.Series = 0,
        window: Optional[int] = None
) -> pd.Series:
    """
    Calculates rolling Beta of a `portfolio`.

    See :func:`~pqr.metrics.beta`.

    Parameters
    ----------
    portfolio
        An allocated portfolio.
    benchmark
        Benchmark for the `portfolio` to beat.
    risk_free_rate
        Indicative rate of guaranteed returns (e.g. US government bond rate).
    window
        Number of observations on a rolling window. If not passed, `window`
        equals to approximate number of periods in a year.
    """

    return _roll(
        portfolio.returns,
        benchmark.returns,
        metric=_beta,
        window=window,
        risk_free_rate=risk_free_rate
    )


def _cumulative_returns(returns: pd.Series) -> pd.Series:
    return (1 + returns).cumprod() - 1


def _total_return(returns: pd.Series) -> int | float:
    return _cumulative_returns(returns).iloc[-1]


def _annual_return(returns: pd.Series) -> int | float:
    annualization_factor = _get_annualization_factor(returns)
    years = len(returns) / annualization_factor
    return (1 + _total_return(returns)) ** (1 / years) - 1


def _annual_volatility(returns: pd.Series) -> int | float:
    annualization_rate = np.sqrt(_get_annualization_factor(returns))
    return _volatility(returns) * annualization_rate


def _mean_return(returns: pd.Series) -> int | float:
    return returns.mean()


def _volatility(returns: pd.Series) -> int | float:
    return returns.std(ddof=1)


def _max_drawdown(returns: pd.Series) -> int | float:
    equity_curve = _cumulative_returns(returns) + 1
    high_water_mark = equity_curve.cummax()
    drawdown = (high_water_mark - equity_curve) / high_water_mark
    return -drawdown[np.isfinite(drawdown)].max()


def _win_rate(returns: pd.Series) -> int | float:
    positive_periods = (returns > 0).sum()
    total_periods = len(returns)
    return positive_periods / total_periods


def _value_at_risk(
        returns: pd.Series,
        confidence_level: int | float = 0.95
) -> int | float:
    return returns.quantile(1 - confidence_level)


def _expected_tail_loss(
        returns: pd.Series,
        confidence_level: int | float = 0.95
):
    var = _value_at_risk(returns, 1 - confidence_level)
    return returns[returns <= var].mean()


def _rachev_ratio(
        returns: pd.Series,
        reward_cutoff: int | float = 0.95,
        risk_cutoff: int | float = 0.05,
        risk_free_rate: int | float | pd.Series = 0
) -> int | float:
    adjusted_returns = _adjust_returns(returns, risk_free_rate)
    reward_var = _value_at_risk(adjusted_returns, 1 - reward_cutoff)
    etr = adjusted_returns[adjusted_returns >= reward_var].mean()
    etl = _expected_tail_loss(adjusted_returns, 1 - risk_cutoff)
    return etr / etl


def _calmar_ratio(returns: pd.Series) -> int | float:
    return _annual_return(returns) / _max_drawdown(returns)


def _sharpe_ratio(
        returns: pd.Series,
        risk_free_rate: int | float | pd.Series = 0
) -> int | float:
    adjusted_returns = _adjust_returns(returns, risk_free_rate)
    annualization_rate = np.sqrt(_get_annualization_factor(adjusted_returns))
    return (_mean_return(adjusted_returns) / _volatility(adjusted_returns) *
            annualization_rate)


def _omega_ratio(
        returns: pd.Series,
        required_return: int | float | pd.Series
) -> int | float:
    adjusted_returns = _adjust_returns(returns, required_return)
    above = adjusted_returns[adjusted_returns > 0].sum()
    under = adjusted_returns[adjusted_returns < 0].sum()
    return above / under


def _sortino_ratio(
        returns: pd.Series,
        minimum_acceptable_return: int | float | pd.Series
) -> int | float:
    adjusted_returns = _adjust_returns(returns, minimum_acceptable_return)
    return _mean_return(adjusted_returns) / _downside_risk(adjusted_returns, 0)


def _downside_risk(
        returns: pd.Series,
        minimum_acceptable_return: int | float | pd.Series = 0
) -> int | float:
    adjusted_returns = _adjust_returns(returns, minimum_acceptable_return)
    annualization_rate = np.sqrt(_get_annualization_factor(adjusted_returns))

    returns_under_mar = np.clip(adjusted_returns, a_min=-np.inf, a_max=0)
    downside_risk_ = np.sqrt((returns_under_mar ** 2).mean())
    return downside_risk_ * annualization_rate


def _mean_excess_return(
        returns: pd.Series,
        benchmark_returns: pd.Series
) -> int | float:
    adjusted_returns = _adjust_returns(returns, benchmark_returns)
    return _mean_return(adjusted_returns)


def _benchmark_correlation(
        returns: pd.Series,
        benchmark_returns: pd.Series
) -> int | float:
    trading_available = returns.index.intersection(benchmark_returns.index)
    return returns.corr(benchmark_returns.loc[trading_available])


def _alpha(
        returns: pd.Series,
        benchmark_returns: pd.Series,
        risk_free_rate: int | float | pd.Series
) -> int | float:
    return _alpha_beta(returns, benchmark_returns, risk_free_rate).iloc[0]


def _beta(
        returns: pd.Series,
        benchmark_returns: pd.Series,
        risk_free_rate: int | float | pd.Series
) -> int | float:
    return _alpha_beta(returns, benchmark_returns, risk_free_rate).iloc[1]


def _alpha_beta(
        returns: pd.Series,
        benchmark_returns: pd.Series,
        risk_free_rate: int | float | pd.Series = 0
) -> pd.Series:
    adjusted_returns = _adjust_returns(returns, risk_free_rate)
    adjusted_benchmark_returns = _adjust_returns(benchmark_returns,
                                                 risk_free_rate)
    trading_available = adjusted_returns.index.intersection(
        adjusted_benchmark_returns.index)
    adjusted_benchmark_returns = adjusted_benchmark_returns[trading_available]
    x = sm_tools.add_constant(adjusted_benchmark_returns)
    est = sm_linear.OLS(adjusted_returns, x).fit()
    return est.params


def _adjust_returns(
        returns: pd.Series,
        benchmark_returns: int | float | pd.Series
):
    if isinstance(benchmark_returns, pd.Series):
        available_index = returns.index.intersection(returns.index)
        benchmark_returns = benchmark_returns[available_index]
    return returns - benchmark_returns


def _roll(
        *returns: pd.Series,
        metric: Callable,
        window: Optional[int] = None,
        **kwargs
) -> pd.Series:
    if window is None:
        window = _get_annualization_factor(returns[0])

    common_index = returns[0].index
    for r in returns:
        common_index = common_index.intersection(r.index)

    values = [np.nan] * (window - 1)
    for i in range(window, len(common_index) + 1):
        idx = common_index[i - window:i]
        rets = [r.loc[idx] for r in returns]
        values.append(metric(*rets, **kwargs))
    return pd.Series(values, index=common_index)


def _get_annualization_factor(data: pd.Series | pd.DataFrame) -> int:
    """
    Extract inferred by pandas frequency of data to annualize metrics. Minimal
    supported frequency of data is daily, maximal - yearly.
    """

    freq_alias_to_num = {
        # yearly
        'BA': 1,
        'A': 1,
        # quarterly
        'BQ': 4,
        'Q': 4,
        # monthly
        'BM': 12,
        'M': 12,
        # weekly
        'W': 52,
        # daily
        'B': 252,
        'D': 252,
    }

    inferred_freq = getattr(data.index, 'inferred_freq', None)
    freq_num = freq_alias_to_num.get(inferred_freq)
    if freq_num is None:
        raise ValueError('periodicity of given data cannot be defined, '
                         'try to resample data')
    return freq_num



"""
This module provides instruments to create benchmarks. Benchmarks treated as
theoretical portfolios, which each investor dreams to beat. Usually there are
already good benchmarks - indices (e.g. S&P 500 or IMOEX). But if for some
reason you cannot find suitable benchmark you can build it from stock universe
(with filters and weights, but without selecting stocks). The benchmark will
include all available (filtered) stock universe in each period  without
transaction costs, but with weighting positions (optionally).

In most cases you need a benchmark just to compare its performance with
performance of a portfolio, but do not forget, that portfolios are also can be
used as benchmarks for calculating metrics/plotting the performance. If you
want to create benchmark with selecting stocks, just construct the portfolio.
"""

from __future__ import annotations

from typing import Optional

import numpy as np
import pandas as pd

import pqr.factors
import pqr.portfolios

__all__ = [
    'Benchmark',
]


class Benchmark:
    """
    Class for benchmarks.

    Parameters
    ----------
    name
        Name of the benchmark.
    """

    name: str
    """Name of the benchmark."""
    returns: pd.Series
    """Periodical returns of the benchmark (non-cumulative)."""

    def __init__(self, name: str = 'benchmark'):
        self.name = name

        self.returns = pd.Series()

    def __repr__(self) -> str:
        return f'Benchmark({repr(self.name)})'

    def __str__(self) -> str:
        return self.name

    def from_index(self, index_values: pd.Series) -> Benchmark:
        """
        Creates benchmark from existing index (e.g. S&P500 or IMOEX).

        Parameters
        ----------
        index_values
            Series of index values. Percentage changes of these values are used
            as returns of the benchmark.
        """

        self.returns = index_values.pct_change()
        self.returns.name = self.name

        return self

    def from_stock_universe(
            self,
            stock_prices: pd.DataFrame,
            mask: Optional[pd.DataFrame] = None,
            weighting_factor: Optional[pqr.factors.Factor] = None
    ) -> Benchmark:
        """
        Creates benchmark from stock universe.

        This type of a benchmark should be used, when there is no relevant
        index to compare with portfolios (e.g. if stock universe is too
        specific or just is severe filtered). In other cases it is highly
        recommended to download data for some existing benchmark, because it is
        really hard to accurately replicate any real stock index.

        To build a benchmark from stock universe you can use filtration and
        weighting, but not picking the stocks. If picking is needed, just
        construct the portfolio, it also can be used as benchmark.

        Parameters
        ----------
        stock_prices
            Prices, representing stock universe. All of available (not np.nan)
            for selecting stocks are picked into the "benchmark-portfolio".
        mask
            Mask to filter stock universe.
        weighting_factor
            Factor, weighting positions (if you want to replicate some
            market-cap weighted benchmark). If not passed, weighs positions
            equally.
        """

        benchmark_portfolio = pqr.portfolios.Portfolio()

        if mask is not None:
            stock_prices[~mask] = np.nan

        picks = ~stock_prices.isna()
        benchmark_portfolio.picks = picks

        if weighting_factor is not None:
            benchmark_portfolio.weigh_by_factor(weighting_factor)
        else:
            benchmark_portfolio.weigh_equally()

        benchmark_portfolio.allocate(stock_prices)

        self.returns = benchmark_portfolio.returns
        self.returns.name = self.name

        return self



"""
This module contains stuff to work with factors data. Factors are drivers of
returns on the stock (and any other) market, explaining the risks of investing
into assets. We assume that a factor can be presented by simple table (pandas
DataFrame), where each row represents a timestamp, each column - a stock
(ticker) and each cell - value of the factor for the stock in the timestamp.

You must be accurate to work with such type of data, because it is very easy to
forget about look-ahead bias and to build very profitable, but unrealistic
factor model or portfolio. There are already included "batteries" to transform
factors with looking, lag and holding periods to avoid it.

Factors can be dynamic or static. Static factor is a factor, for which is not
necessary to calculate the change, its value can be compared across the stocks
straightly (e.g. P/E). Dynamic factor is a factor, which values must be
recalculated to get the percentage changes to work with them (e.g. Momentum).
Start of testing for dynamic factors will be 1 period later to avoid look-ahead
bias.

Factors also have one more characteristic - whether values of factor is better
to be bigger or lower. For example, we usually want to pick into portfolio
stocks with low P/E, hence, this factor is "lower better", whereas stocks with
high ROA are usually more preferable, hence, this factor is "bigger better".
This characteristic affects building wml-portfolios for factor model and
weighting and scaling (leveraging) of positions in a portfolio.
"""

from __future__ import annotations

from typing import Literal

import numpy as np
import pandas as pd

__all__ = [
    'Factor',
]


class Factor:
    """
    Class for factors, represented by matrix of numeric values.

    Parameters
    ----------
    data
        Matrix of factor values.
    """

    data: pd.DataFrame
    """Factor values."""

    def __init__(self, data: pd.DataFrame):
        self.data = data.copy()

    def transform(
            self,
            looking_back_period: int = 1,
            method: Literal['static', 'dynamic'] = 'static',
            lag_period: int = 0,
            holding_period: int = 1
    ) -> Factor:
        """
        Transforms `factor` into appropriate for decision-making format.

        This function helps to preprocess `factor` before using it to make a
        portfolio and prevents from look-ahead bias.

        Parameters
        ----------
        looking_back_period
            Looking back period for `factor`.
        method
            Whether absolute values of `factor` are used to make decision or
            their percentage changes.
        lag_period
            Delaying period to react on `factor`.
        holding_period
            Number of periods to hold each `factor` value.
        """

        return (self.look_back(looking_back_period, method)
                .lag(lag_period)
                .hold(holding_period))

    def look_back(
            self,
            period: int = 1,
            method: Literal['static', 'dynamic'] = 'static',
    ) -> Factor:
        """
        Looks back on `factor` for `period` periods.

        If `method` is dynamic:
            calculates percentage changes with looking back by `period`
            periods, then values are lagged for 1 period (because in period
            t(0) we can know percentage change from period t(-looking_period)
            only at the end of t(0), so it is needed to avoid look-ahead bias).

        If `method` is static:
            all values are shifted for `period`.

        Parameters
        ----------
        period
            Looking back period for `factor`.
        method
            Whether absolute values of `factor` are used to make decision or
            their percentage changes.
        """

        if method == 'dynamic':
            self.data = self.data.pct_change(period)
            self.lag(1)
        else:  # method = 'static'
            self.data = self.data.shift(period)

        self.data = self.data.iloc[period:]

        return self

    def lag(self, period: int = 0) -> Factor:
        """
        Simply shifts the `factor` for `period` periods.

        Can be used to simulate delayed reaction on `factor` values.

        Parameters
        ----------
        period
            Delaying period to react on `factor`.
        """

        self.data = self.data.shift(period).iloc[period:]

        return self

    def hold(self, period: int = 1) -> Factor:
        """
        Fills forward row-wise `factor` with the periodicity of `period`.

        Can be used to simulate periodical updates of `factor`.

        Parameters
        ----------
        period
            Number of periods to hold each `factor` value.
        """

        if period == 1:
            return self

        all_periods = np.zeros(len(self.data), dtype=int)
        update_periods = np.arange(len(self.data), step=period)
        all_periods[update_periods] = update_periods
        update_mask = np.maximum.accumulate(all_periods[:, np.newaxis], axis=0)

        self.data = pd.DataFrame(
            np.take_along_axis(self.data.values, update_mask, axis=0),
            index=self.data.index, columns=self.data.columns)

        return self

    def prefilter(
            self,
            mask: pd.DataFrame
    ) -> Factor:
        """
        Filters the `factor` by given `mask`.

        Simply deletes (replaces with np.nan) cells, where the `mask` equals
        to False.

        Parameters
        ----------
        mask
            Matrix of True/False, where True means that a value should remain
            in `factor` and False - that a value should be deleted.
        """

        self.data[~mask] = np.nan

        return self



"""
This module contains all necessary functionality to create factor models.
Factor model is the set of portfolios, covering all stock universe in each
period. It is used to test whether any factor is present on the market or not,
that is why full coverage of stock universe is important.

Usually factor models are built for single-factors with the quantiles method.
So, only this way to fit factor models is natively supported, but if you need
to make full coverage of stock universe with time-series method or top-n
portfolios, you can make it by yourself, using portfolios module.

Users of factor models are often interested in performance of the
wml-portfolio, and this functionality is also supported.

Moreover, you can calibrate factor model parameters, using simple bruteforce
method, provided by grid_search() function.
"""

from __future__ import annotations

from typing import Tuple, List, Optional, Iterable, Dict, Literal

import numpy as np
import pandas as pd

import pqr.benchmarks
import pqr.factors
import pqr.portfolios

__all__ = [
    'fit_factor_model',
    'calculate_portfolios_summary_stats',
    'factor_model_tear_sheet',
    'grid_search',
]


def fit_factor_model(
        stock_prices: pd.DataFrame,
        factor: pqr.factors.Factor,
        weighting_factor: Optional[pqr.factors.Factor] = None,
        balance: Optional[int | float] = None,
        fee_rate: int | float = 0,
        quantiles: int = 3,
        add_wml: bool = False,
        is_bigger_better: bool = True
) -> List[pqr.portfolios.AbstractPortfolio]:
    """
    Fits factor model with quantile-method.

    Creates `quantiles` portfolios, covering all stock universe and
    (optionally) add wml-portfolio.

    Parameters
    ----------
    stock_prices
        Prices, representing stock universe.
    factor
        Factor to pick stocks into the portfolio.
    weighting_factor
        Factor to weigh picks.
    balance
        Initial balance of the portfolio.
    fee_rate
        Indicative commission rate for every deal.
    quantiles
        Number of portfolios to build for covering stock universe by quantiles.
    add_wml
        Whether to also create wml-portfolio or not.
    is_bigger_better
        Whether portfolio with highest quantiles will be "winners" for
        wml-portfolio or with lowest ones.
    """

    quantiles_ = _split_quantiles(quantiles)

    portfolios = []
    for q in quantiles_:
        portfolio = pqr.portfolios.Portfolio('q({:.2f}, {:.2f})'.format(*q))
        portfolio.pick_stocks_by_factor(factor, q)

        if weighting_factor is not None:
            portfolio.weigh_by_factor(factor)
        else:
            portfolio.weigh_equally()

        portfolio.allocate(stock_prices, balance, fee_rate)

        portfolios.append(portfolio)

    if add_wml:
        if is_bigger_better:
            wml = pqr.portfolios.WmlPortfolio(portfolios[-1],
                                              portfolios[0])
        else:
            wml = pqr.portfolios.WmlPortfolio(portfolios[0],
                                              portfolios[-1])
        portfolios.append(wml)

    return portfolios


def calculate_portfolios_summary_stats(
        *portfolios: pqr.portfolios.AbstractPortfolio,
        benchmark: pqr.benchmarks.Benchmark
) -> pd.DataFrame:
    """
    Calculates portfolios summary statistics and gather them into a table.

    See function pqr.metrics.summary().

    Parameters
    ----------
    portfolios
        Portfolios, for which summary stats are calculated.
    benchmark
        Benchmark to compute some metrics.
    """

    stats = pd.DataFrame(
        [pqr.metrics.summary(p, benchmark) for p in portfolios]).T.round(2)
    return stats


def factor_model_tear_sheet(
        *portfolios: pqr.portfolios.AbstractPortfolio,
        benchmark: pqr.benchmarks.Benchmark,
) -> pd.DataFrame:
    """
    Shows the performance assessment of a factor model' portfolios.

    For now:

    * shows summary stats table
    * plots cumulative returns

    Parameters
    ----------
    portfolios
        Portfolios, included into the factor model.
    benchmark
        Benchmark to compute some metrics.
    """

    stats = calculate_portfolios_summary_stats(*portfolios,
                                               benchmark=benchmark)
    pqr.visualization.plot_cumulative_returns(*portfolios, benchmark=benchmark)
    return stats


def grid_search(
        stock_prices: pd.DataFrame,
        factor: pd.DataFrame,
        looking_back_periods: Iterable[int],
        method: Literal['static', 'dynamic'],
        lag_periods: Iterable[int],
        holding_periods: Iterable[int],
        benchmark: pqr.benchmarks.Benchmark,
        mask: Optional[pd.DataFrame] = None,
        **kwargs
) -> Dict[Tuple[int, int, int], pd.DataFrame]:
    """
    Fits a grid of factor models.

    Can be used to find the best parameters or just as fast alias to build a
    lot of models with different parameters.

    Parameters
    ----------
    stock_prices
        Prices, representing stock universe.
    factor
        Factor, used to pick stocks from (filtered) stock universe.
    looking_back_periods
        Looking back periods for `factor` to be tested.
    method
        Whether absolute values of `factor` are used to make decision or
        their percentage changes.
    lag_periods
        Lag periods for `factor` to be tested.
    holding_periods
        Holding periods for `factor` to be tested.
    benchmark
        Benchmark to compare with portfolios and calculate some metrics.
    mask
        Matrix of True/False, where True means that a value should remain
        in `factor` and False - that a value should be deleted.
    **kwargs
        Keyword arguments for fitting factor models. See fit_factor_model().
    """

    results = {}
    for looking in looking_back_periods:
        for lag in lag_periods:
            for holding in holding_periods:
                transformed_factor = pqr.factors.Factor(factor)
                transformed_factor.transform(looking, method, lag, holding)
                if mask is not None:
                    transformed_factor.prefilter(mask)
                portfolios = fit_factor_model(stock_prices, transformed_factor,
                                              **kwargs)
                results[(looking, lag, holding)] = (
                    calculate_portfolios_summary_stats(*portfolios,
                                                       benchmark=benchmark)
                )
    return results


def _split_quantiles(n: int) -> List[Tuple[int | float, int | float]]:
    """
    Creates `n` quantiles, covering all range from 0 to 1.
    """

    quantile_pairs = np.take(np.linspace(0, 1, n + 1),
                             np.arange(n * 2).reshape((n, -1)) -
                             np.indices((n, 2))[0])
    return [tuple(pair) for pair in quantile_pairs]



import setuptools


with open('README.md') as f:
    readme = f.read()


setuptools.setup(
    name='pqr',
    version='0.1.1',
    description='Library for testing factor strategies',
    long_description=readme,
    long_description_content_type='text/markdown',
    packages=setuptools.find_packages(),
    install_requires=[
        'pandas<=1.3.0',
        'matplotlib<=3.4.2',
        'numpy<=1.21.0',
        'statsmodels<=0.12.2',
    ],
    author='eura17, atomtosov',
    license='MIT',
    license_file='LICENSE',
    platforms='any',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Office/Business :: Financial',
        'Topic :: Scientific/Engineering :: Information Analysis',
    ],
    test_suite='pytest',
    python_requires='>=3.7',
    project_urls={
            'Bug Tracker': 'https://github.com/atomtosov/pqr/issues',
            'Documentation': 'https://pqr.readthedocs.io/en/latest/index.html',
            'Source Code': 'https://github.com/atomtosov/pqr/tree/dev',
        },
)
